# Copyright (C) 2021, QuantStack
# SPDX-License-Identifier: BSD-3-Clause

from rich.console import Console

boa_config = None


class BoaConfig:
    console = Console()
    json: bool = False

    def __init__(self, args=None):
        if args and getattr(args, "json", False):
            self.console.quiet = True
            self.json = True


def init_global_config(args=None):
    global boa_config
    boa_config = BoaConfig(args)


if not boa_config:
    init_global_config()



# Copyright (C) 2021, QuantStack
# SPDX-License-Identifier: BSD-3-Clause

import os
import re
import time
from os.path import exists, isdir, join
from subprocess import CalledProcessError

import yaml
from bs4 import UnicodeDammit

from conda.gateways.disk.create import mkdir_p
from conda_build import utils
from conda_build.utils import check_output_env, get_logger

from boa import __version__ as boa_version


def get_repository_info(recipe_path):
    """This tries to get information about where a recipe came from.  This is different
    from the source - you can have a recipe in svn that gets source via git."""
    try:
        if exists(join(recipe_path, ".git")):
            origin = check_output_env(
                ["git", "config", "--get", "remote.origin.url"], cwd=recipe_path
            )
            rev = check_output_env(["git", "rev-parse", "HEAD"], cwd=recipe_path)
            return "Origin {}, commit {}".format(origin, rev)
        elif isdir(join(recipe_path, ".hg")):
            origin = check_output_env(["hg", "paths", "default"], cwd=recipe_path)
            rev = check_output_env(["hg", "id"], cwd=recipe_path).split()[0]
            return "Origin {}, commit {}".format(origin, rev)
        elif isdir(join(recipe_path, ".svn")):
            info = check_output_env(["svn", "info"], cwd=recipe_path)
            info = info.decode(
                "utf-8"
            )  # Py3 returns a byte string, but re needs unicode or str.
            server = re.search("Repository Root: (.*)$", info, flags=re.M).group(1)
            revision = re.search("Revision: (.*)$", info, flags=re.M).group(1)
            return "{}, Revision {}".format(server, revision)
        else:
            return "{}, last modified {}".format(
                recipe_path,
                time.ctime(os.path.getmtime(join(recipe_path, "recipe.yaml"))),
            )
    except CalledProcessError:
        get_logger(__name__).debug("Failed to checkout source in " + recipe_path)
        return "{}, last modified {}".format(
            recipe_path, time.ctime(os.path.getmtime(join(recipe_path, "recipe.yaml")))
        )


def _copy_top_level_recipe(path, config, dest_dir, destination_subdir=None):
    files = utils.rec_glob(path, "*")
    file_paths = sorted([f.replace(path + os.sep, "") for f in files])

    # when this actually has a value, we're copying the top-level recipe into a subdirectory,
    #    so that we have record of what parent recipe produced subpackages.
    if destination_subdir:
        dest_dir = join(dest_dir, destination_subdir)
    else:
        # exclude recipe.yaml because the json dictionary captures its content
        file_paths = [
            f
            for f in file_paths
            if not (f == "recipe.yaml" or f == "conda_build_config.yaml")
        ]
    file_paths = utils.filter_files(file_paths, path)
    for f in file_paths:
        utils.copy_into(
            join(path, f),
            join(dest_dir, f),
            timeout=config.timeout,
            locking=config.locking,
            clobber=True,
        )


def _copy_output_recipe(m, dest_dir):
    _copy_top_level_recipe(m.path, m.config, dest_dir, "parent")

    this_output = m.get_rendered_output(m.name()) or {}
    install_script = this_output.get("script")
    build_inputs = []
    inputs = [install_script] + build_inputs
    file_paths = [script for script in inputs if script]
    file_paths = utils.filter_files(file_paths, m.path)

    for f in file_paths:
        utils.copy_into(
            join(m.path, f),
            join(dest_dir, f),
            timeout=m.config.timeout,
            locking=m.config.locking,
            clobber=True,
        )


def output_yaml(metadata, filename=None, suppress_outputs=False):
    local_metadata = metadata.copy()
    if (
        suppress_outputs
        and local_metadata.is_output
        and "outputs" in local_metadata.meta
    ):
        del local_metadata.meta["outputs"]
    output = yaml.dump((local_metadata.meta), default_flow_style=False, indent=4)
    if filename:
        if any(sep in filename for sep in ("\\", "/")):
            mkdir_p(os.path.dirname(filename))
        with open(filename, "w") as f:
            f.write(output)
        return "Wrote yaml to %s" % filename
    else:
        return output


def copy_recipe(m):
    if m.config.include_recipe and m.include_recipe():
        # store the rendered recipe.yaml file, plus information about where it came from
        #    and what version of conda-build created it
        recipe_dir = join(m.config.info_dir, "recipe")
        mkdir_p(recipe_dir)

        original_recipe = ""

        if m.is_output:
            _copy_output_recipe(m, recipe_dir)
        else:
            _copy_top_level_recipe(m.path, m.config, recipe_dir)
            original_recipe = m.meta_path

        output_metadata = m.copy()
        # hard code the build string, so that tests don't get it mixed up
        build = output_metadata.meta.get("build", {})
        build["string"] = output_metadata.build_id()
        output_metadata.meta["build"] = build

        # just for lack of confusion, don't show outputs in final rendered recipes
        if "outputs" in output_metadata.meta:
            del output_metadata.meta["outputs"]
        if "parent_recipe" in output_metadata.meta.get("extra", {}):
            del output_metadata.meta["extra"]["parent_recipe"]

        utils.sort_list_in_nested_structure(
            output_metadata.meta, ("build/script", "test/commands")
        )

        rendered = output_yaml(output_metadata)

        if original_recipe:
            with open(original_recipe, "rb") as f:
                original_recipe_text = UnicodeDammit(f.read()).unicode_markup

        if not original_recipe or not original_recipe_text == rendered:
            with open(join(recipe_dir, "recipe.yaml"), "w") as f:
                f.write("# This file created by boa {}\n".format(boa_version))
                if original_recipe:
                    f.write("# recipe.yaml template originally from:\n")
                    f.write("# " + get_repository_info(m.path) + "\n")
                f.write("# ------------------------------------------------\n\n")
                f.write(rendered)
            if original_recipe:
                utils.copy_into(
                    original_recipe,
                    os.path.join(recipe_dir, "recipe.yaml.template"),
                    timeout=m.config.timeout,
                    locking=m.config.locking,
                    clobber=True,
                )

        # dump the full variant in use for this package to the recipe folder
        with open(os.path.join(recipe_dir, "conda_build_config.yaml"), "w") as f:
            yaml.dump(m.config.variant, f)



# Copyright (C) 2021, QuantStack
# SPDX-License-Identifier: BSD-3-Clause

from glob import glob
import os
from pathlib import Path
from math import log
from mamba.mamba_api import transmute as mamba_transmute
from joblib import Parallel, delayed

from rich.console import Console

console = Console()

unit_list = list(zip(["bytes", "kB", "MB", "GB", "TB", "PB"], [0, 0, 1, 2, 2, 2]))


def sizeof_fmt(num):
    """Human friendly file size"""
    if num > 1:
        exponent = min(int(log(num, 1024)), len(unit_list) - 1)
        quotient = float(num) / 1024 ** exponent
        unit, num_decimals = unit_list[exponent]
        format_string = "{:.%sf} {}" % (num_decimals)
        return format_string.format(quotient, unit)
    if num == 0:
        return "0 bytes"
    if num == 1:
        return "1 byte"


def transmute_task(f, args):
    filename = os.path.basename(f)
    outpath = os.path.abspath(args.output_directory)

    if f.endswith(".tar.bz2"):
        filename = filename[:-8]
        outfile = os.path.join(outpath, filename + ".conda")
    elif f.endswith(".conda"):
        filename = filename[:-6]
        outfile = os.path.join(outpath, filename + ".tar.bz2")
    else:
        console.print("[bold red]Transmute can only handle .tar.bz2 and .conda formats")

    console.print(f"Processing {filename}")
    mamba_transmute(f, outfile, args.compression_level)

    stat_before = Path(f).stat()
    stat_after = Path(outfile).stat()

    saved_percent = 1.0 - (stat_after.st_size / stat_before.st_size)
    color = "[bold green]" if saved_percent > 0 else "[bold red]"

    return filename, outfile, stat_before, stat_after, saved_percent, color


def main(args):
    # from mamba.mamba_api import Context
    # api_ctx = Context()
    # api_ctx.set_verbosity(1)

    # from mamba.mamba_api import Context
    # api_ctx = Context()
    # api_ctx.set_verbosity(1)

    files = args.files
    final_files = []

    if not os.path.exists(args.output_directory):
        Path(args.output_directory).mkdir(parents=True, exist_ok=True)

    for f in files:
        final_files += [os.path.abspath(fx) for fx in glob(f)]

    logs = Parallel(n_jobs=args.num_jobs)(
        delayed(transmute_task)(f, args) for f in final_files
    )

    for filename, outfile, stat_before, stat_after, saved_percent, color in logs:
        console.print(f"\nConverting [bold]{filename}")
        console.print(f"Done: [bold]{outfile}")
        console.print(f"   Before    : {sizeof_fmt(stat_before.st_size)}")
        console.print(f"   After     : {sizeof_fmt(stat_after.st_size)}")
        console.print(f"   Difference: {color}{saved_percent * 100:.2f}%")



# Copyright (C) 2021, QuantStack
# SPDX-License-Identifier: BSD-3-Clause

import sys
import argparse

from boa.core.config import init_global_config

from mamba.utils import init_api_context

banner = r"""
           _
          | |__   ___   __ _
          | '_ \ / _ \ / _` |
          | |_) | (_) | (_| |
          |_.__/ \___/ \__,_|
"""


def main(config=None):

    parser = argparse.ArgumentParser(
        description="Boa, the fast, mamba powered-build tool for conda packages."
    )

    subparsers = parser.add_subparsers(help="sub-command help", dest="command")
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument("--recipe-dir", type=str)
    parent_parser.add_argument("target", type=str, default="")
    parent_parser.add_argument("--features", type=str)
    parent_parser.add_argument("--offline", action="store_true")
    parent_parser.add_argument("--target-platform", type=str)
    parent_parser.add_argument("--json", action="store_true")

    variant_parser = argparse.ArgumentParser(add_help=False)
    variant_parser.add_argument(
        "-m",
        "--variant-config-files",
        action="append",
        help="""Additional variant config files to add.  These yaml files can contain
        keys such as `c_compiler` and `target_platform` to form a build matrix.""",
    )

    subparsers.add_parser(
        "render", parents=[parent_parser, variant_parser], help="render a recipe"
    )
    subparsers.add_parser(
        "convert",
        parents=[parent_parser],
        help="convert recipe.yaml to old-style meta.yaml",
    )
    subparsers.add_parser(
        "validate", parents=[parent_parser], help="Validate recipe.yaml",
    )

    build_parser = argparse.ArgumentParser(add_help=False)
    build_parser.add_argument(
        "-i",
        "--interactive",
        action="store_true",
        help="Use interactive mode if build fails",
    )
    build_parser.add_argument(
        "--skip-existing", nargs="?", default="default", const="yes",
    )
    build_parser.add_argument(
        "--no-test",
        action="store_true",
        dest="notest",
        help="Do not test the package.",
    )
    build_parser.add_argument(
        "--continue-on-failure",
        action="store_true",
        help="Continue building remaining recipes if a recipe fails.",
    )

    subparsers.add_parser(
        "build",
        parents=[parent_parser, build_parser, variant_parser],
        help="build a recipe",
    )

    transmute_parser = subparsers.add_parser(
        "transmute",
        parents=(),
        help="transmute one or many tar.bz2 packages into a conda packages (or vice versa!)",
    )
    transmute_parser.add_argument("files", type=str, nargs="+")
    transmute_parser.add_argument("-o", "--output-directory", type=str, default=".")
    transmute_parser.add_argument("-c", "--compression-level", type=int, default=22)
    transmute_parser.add_argument(
        "-n_jobs",
        "--num_jobs",
        type=int,
        default=1,
        help="the number of parallel processing elements",
    )

    args = parser.parse_args()

    command = args.command

    init_api_context()
    init_global_config(args)

    from boa.core.run_build import run_build
    from boa.cli import convert
    from boa.cli import transmute
    from boa.cli import validate

    if command == "convert":
        convert.main(args.target)
        exit()

    if command == "validate":
        validate.main(args.target)
        exit()

    if command == "transmute":
        transmute.main(args)
        exit()

    from boa.core.config import boa_config

    boa_config.console.print(banner)

    if command == "build" or command == "render":
        run_build(args)

    if not command:
        parser.print_help(sys.stdout)


if __name__ == "__main__":
    main()



# Copyright (C) 2021, QuantStack
# SPDX-License-Identifier: BSD-3-Clause

from boa.core.solver import get_solver
import copy
import re
import json
from pathlib import Path
import sys

from dataclasses import dataclass
from typing import Tuple

import rich
from rich.table import Table
from rich.padding import Padding


# TODO remove
from conda.base.context import context
from conda_build.config import get_or_merge_config
from conda_build.utils import apply_pin_expressions
from conda.models.channel import Channel as CondaChannel
from conda_build.metadata import eval_selector, ns_cfg

from boa.core.config import boa_config

console = boa_config.console


@dataclass
class CondaBuildSpec:
    name: str
    raw: str
    splitted: Tuple[str]
    is_pin: bool = False
    is_pin_compatible: bool = False
    is_compiler: bool = False
    is_transitive_dependency: bool = False
    channel: str = ""
    # final: String

    from_run_export: bool = False
    from_pinnings: bool = False

    def __init__(self, ms):
        self.raw = ms
        self.splitted = ms.split()
        self.name = self.splitted[0]
        if len(self.splitted) > 1:
            self.is_pin = self.splitted[1].startswith("PIN_")
            self.is_pin_compatible = self.splitted[1].startswith("PIN_COMPATIBLE")
            self.is_compiler = self.splitted[0].startswith("COMPILER_")

        self.is_simple = len(self.splitted) == 1
        self.final = self.raw

        if self.is_pin_compatible:
            self.final[len("PIN_COMPATIBLE") + 1 : -1]

    @property
    def final_name(self):
        return self.final.split(" ")[0]

    def loosen_spec(self):
        if self.is_compiler or self.is_pin:
            return

        if len(self.splitted) == 1:
            return

        if re.search(r"[^0-9\.]+", self.splitted[1]) is not None:
            return

        dot_c = self.splitted[1].count(".")

        app = "*" if dot_c >= 2 else ".*"

        if len(self.splitted) == 3:
            self.final = (
                f"{self.splitted[0]} {self.splitted[1]}{app} {self.splitted[2]}"
            )
        else:
            self.final = f"{self.splitted[0]} {self.splitted[1]}{app}"

    def __repr__(self):
        self.loosen_spec()
        return self.final

    def eval_pin_subpackage(self, all_outputs):
        if len(self.splitted) <= 1 or not self.splitted[1].startswith("PIN_SUBPACKAGE"):
            return
        pkg_name = self.name
        max_pin, exact = self.splitted[1][len("PIN_SUBPACKAGE") + 1 : -1].split(",")
        exact = exact == "True"
        output = None
        # TODO are we pinning the right version if building multiple variants?!
        for o in all_outputs:
            if o.name == pkg_name:
                output = o
                break

        if not output:
            raise RuntimeError(f"Could not find output with name {pkg_name}")
        version = output.version
        build_string = output.final_build_id

        if exact:
            self.final = f"{pkg_name} {version} {build_string}"
        else:
            version_parts = version.split(".")
            count_pin = max_pin.count(".")
            version_pin = ".".join(version_parts[: count_pin + 1])
            version_pin += ".*"
            self.final = f"{pkg_name} {version_pin}"

    def eval_pin_compatible(self, build, host):
        if len(self.splitted) <= 1:
            return

        lower_bound, upper_bound, min_pin, max_pin, exact = self.splitted[1][
            len("PIN_COMPATIBLE") + 1 : -1
        ].split(",")
        if lower_bound == "None":
            lower_bound = None
        if upper_bound == "None":
            upper_bound = None
        exact = exact == "True"

        versions = {b.name: b for b in build}
        versions.update({h.name: h for h in host})

        compatibility = None
        if versions:
            if exact and versions.get(self.name):
                compatibility = " ".join(versions[self.name].final_version)
            else:
                version = lower_bound or versions.get(self.name).final_version[0]
                if version:
                    if upper_bound:
                        if min_pin or lower_bound:
                            compatibility = ">=" + str(version) + ","
                        compatibility += "<{upper_bound}".format(
                            upper_bound=upper_bound
                        )
                    else:
                        compatibility = apply_pin_expressions(version, min_pin, max_pin)

        self.final = (
            " ".join((self.name, compatibility))
            if compatibility is not None
            else self.name
        )


class Output:
    def __init__(
        self, d, config, parent=None, conda_build_config=None, selected_features=None
    ):
        if parent is None:
            parent = {}
        if selected_features is None:
            selected_features = {}
        self.data = d
        self.data["source"] = d.get("source", parent.get("source", {}))
        self.config = config
        self.conda_build_config = conda_build_config or {}
        self.name = d["package"]["name"]
        self.version = d["package"]["version"]
        self.build_string = d["package"].get("build_string")
        self.build_number = d["build"].get("number", 0)
        self.noarch = d["build"].get("noarch", False)
        self.is_first = False
        self.sections = {}

        def set_section(sname):
            self.sections[sname] = {}
            self.sections[sname].update(parent.get(sname, {}))
            self.sections[sname].update(d.get(sname, {}))

        set_section("build")
        set_section("package")
        set_section("app")
        set_section("extra")
        set_section("test")

        self.sections["files"] = d.get("files")
        self.sections["source"] = self.data.get("source", {})
        if hasattr(self.sections["source"], "keys"):
            self.sections["source"] = [self.sections["source"]]

        self.sections["features"] = parent.get("features", [])

        self.feature_map = {f["name"]: f for f in self.sections.get("features", [])}
        for fname, feat in self.feature_map.items():
            activated = feat.get("default", False)
            if fname in selected_features:
                activated = selected_features[fname]

            feat["activated"] = activated

        if self.feature_map.get("static") and self.feature_map["static"]["activated"]:
            self.name += "-static"

        if len(self.feature_map):
            table = Table()
            table.title = "Activated Features"
            table.add_column("Feature")
            table.add_column("State")
            for feature in self.feature_map:
                if self.feature_map[feature]["activated"]:
                    table.add_row(feature, "[green]ON[/green]")
                else:
                    table.add_row(feature, "[red]OFF[/red]")

            console.print(table)

        self.requirements = copy.copy(d.get("requirements", {}))
        for f in self.feature_map.values():
            if f["activated"]:
                if not f.get("requirements"):
                    continue
                for i in ["build", "host", "run", "run_constrained"]:
                    base_req = self.requirements.get(i, [])
                    feat_req = f["requirements"].get(i, [])
                    base_req += feat_req
                    if len(base_req):
                        self.requirements[i] = base_req

        self.transactions = {}

        self.parent = parent

        for section in ("build", "host", "run", "run_constrained"):
            self.requirements[section] = [
                CondaBuildSpec(r) for r in (self.requirements.get(section) or [])
            ]
        self.sections["requirements"] = self.requirements

        # handle strong and weak run exports
        run_exports = []
        for el in self.sections["build"].get("run_exports", []):
            if type(el) is str:
                run_exports.append(CondaBuildSpec(el))
            else:
                raise RuntimeError("no strong run exports supported as of now.")
                # sub_run_exports = []
                # for key, val in el:
                #     for x in val:
                #         sub_run_exports.append(CondaBuildSpec(x))
                #     run_exports.append({})
        if run_exports:
            self.sections["build"]["run_exports"] = run_exports

    def skip(self):
        skips = self.sections["build"].get("skip", [])
        skip_reasons = []
        for x in skips:
            if eval_selector(x, ns_cfg(self.config), []):
                skip_reasons.append(x)
        if len(skip_reasons):
            console.print(
                f"[green]Skipping {self.name} {' | '.join(self.differentiating_variant)} because of[/green]\n"
                + "\n".join([f"  - {x}" for x in skip_reasons])
            )
        return len(skip_reasons) != 0

    def all_requirements(self):
        requirements = (
            self.requirements.get("build")
            + self.requirements.get("host")
            + self.requirements.get("run")
        )
        return requirements

    def apply_variant(self, variant, differentiating_keys=()):
        copied = copy.deepcopy(self)

        copied.variant = variant
        for idx, r in enumerate(self.requirements["build"]):
            vname = r.name.replace("-", "_")
            if vname in variant:
                copied.requirements["build"][idx] = CondaBuildSpec(
                    r.name + " " + variant[vname]
                )
                copied.requirements["build"][idx].from_pinnings = True
        for idx, r in enumerate(self.requirements["host"]):
            vname = r.name.replace("-", "_")
            if vname in variant:
                copied.requirements["host"][idx] = CondaBuildSpec(
                    r.name + " " + variant[vname]
                )
                copied.requirements["host"][idx].from_pinnings = True

        # todo figure out if we should pin like that in the run reqs as well?
        # for idx, r in enumerate(self.requirements["run"]):
        #     vname = r.name.replace("-", "_")
        #     if vname in variant:
        #         copied.requirements["run"][idx] = CondaBuildSpec(
        #             r.name + " " + variant[vname]
        #         )
        #         copied.requirements["run"][idx].from_pinnings = True

        # insert compiler_cxx, compiler_c and compiler_fortran
        for idx, r in enumerate(self.requirements["build"]):
            if r.name.startswith("COMPILER_"):
                lang = r.splitted[1].lower()
                if variant.get(lang + "_compiler"):
                    compiler = (
                        f"{variant[lang + '_compiler']}_{variant['target_platform']}"
                    )
                if variant.get(lang + "_compiler_version"):
                    version = variant[lang + "_compiler_version"]
                    copied.requirements["build"][idx].final = f"{compiler} {version}*"
                else:
                    copied.requirements["build"][idx].final = f"{compiler}"
                copied.requirements["build"][idx].from_pinnings = True

        for r in self.requirements["host"]:
            if r.name.startswith("COMPILER_"):
                raise RuntimeError("Compiler should be in build section")

        copied.config = get_or_merge_config(self.config, variant=variant)

        copied.differentiating_variant = []
        for k in differentiating_keys:
            copied.differentiating_variant.append(variant[k])

        return copied

    def to_json(self):
        res = {
            "name": self.name,
            "version": self.version,
            "build_number": self.build_number,
            "source": self.data["source"],
            "noarch": self.noarch,
        }

        res["differentiating_variant"] = self.differentiating_variant
        res["variant"] = self.variant

        def specs_to_dict(specs):
            ret = []
            for s in specs:
                attrs = []
                if s.is_pin:
                    attrs.append("is_pin")
                if s.from_run_export:
                    attrs.append("from_run_export")
                if s.from_pinnings:
                    attrs.append("from_pinnings")

                sdict = {
                    "spec": s.raw,
                    "name": s.final_name,
                    "attrs": attrs,
                    "resolved": {},
                }
                if hasattr(s, "final_version"):
                    sdict["resolved"]["final_version"] = s.final_version

                ret.append(sdict)
            return ret

        res["requirements"] = {
            "build": specs_to_dict(self.requirements["build"]),
            "host": specs_to_dict(self.requirements["host"]),
            "run": specs_to_dict(self.requirements["run"]),
        }

        return res

    def __rich__(self):
        table = Table(box=rich.box.MINIMAL_DOUBLE_HEAD)
        s = f"Output: {self.name} {self.version} BN: {self.build_number}\n"
        if hasattr(self, "differentiating_variant"):
            short_v = " ".join([val for val in self.differentiating_variant])
            s += f"Variant: {short_v}\n"
        s += "Build:\n"
        table.title = s
        table.add_column("Dependency")
        table.add_column("Version requirement")
        table.add_column("Selected")
        table.add_column("Build")
        table.add_column("Channel")

        def spec_format(x):
            version, fv = " ", " "
            channel = CondaChannel.from_url(x.channel).name

            if (
                x.channel.startswith("file://")
                and context.local_build_root in x.channel
            ):
                channel = "local"

            if len(x.final.split(" ")) > 1:
                version = " ".join(r.final.split(" ")[1:])
            if hasattr(x, "final_version"):
                fv = x.final_version
            color = "white"
            if x.from_run_export:
                color = "blue"
            if x.from_pinnings:
                color = "green"
            if x.is_transitive_dependency:
                table.add_row(
                    f"{r.final_name}", "", f"{fv[0]}", f"{fv[1]}", f"{channel}"
                )
                return

            if x.is_pin:
                if x.is_pin_compatible:
                    version = "PC " + version
                else:
                    version = "PS " + version
                color = "cyan"

            if len(fv) >= 2:
                table.add_row(
                    f"[bold white]{r.final_name}[/bold white]",
                    f"[{color}]{version}[/{color}]",
                    f"{fv[0]}",
                    f"{fv[1]}",
                    f"{channel}",
                )
            else:
                table.add_row(
                    f"[bold white]{r.final_name}[/bold white]",
                    f"[{color}]{version}[/{color}]",
                    f"{fv[0]}",
                    "",
                    f"{channel}",
                )

        def add_header(header, head=False):
            p = Padding("", (0, 0), style="black")
            if head:
                pns = Padding("", (0, 0), style="black")
                table.add_row(pns, pns, pns, pns, pns)
            table.add_row(Padding(header, (0, 0), style="bold yellow"), p, p, p, p)

        if self.requirements["build"]:
            add_header("Build")
            for r in self.requirements["build"]:
                spec_format(r)
        if self.requirements["host"]:
            add_header("Host", True)
            for r in self.requirements["host"]:
                spec_format(r)
        if self.requirements["run"]:
            add_header("Run", True)
            for r in self.requirements["run"]:
                spec_format(r)
        if self.requirements["run_constrained"]:
            add_header("Run Constraints", True)
            for r in self.requirements["run_constrained"]:
                spec_format(r)
        return table

    def __repr__(self):
        s = f"Output: {self.name} {self.version} BN: {self.build_number}\n"
        if hasattr(self, "differentiating_variant"):
            short_v = " ".join([val for val in self.differentiating_variant])
            s += f"Variant: {short_v}\n"
        s += "Build:\n"

        def spec_format(x):
            version, fv = " ", " "
            if len(x.final.split(" ")) > 1:
                version = " ".join(r.final.split(" ")[1:])
            if hasattr(x, "final_version"):
                fv = x.final_version
            color = "white"
            if x.from_run_export:
                color = "blue"
            if x.from_pinnings:
                color = "green"
            if x.is_transitive_dependency:
                return f" - {r.final_name:<51} {fv[0]:<10} {fv[1]:<10}\n"
            if x.is_pin:
                if x.is_pin_compatible:
                    version = "PC " + version
                else:
                    version = "PS " + version
                color = "cyan"

            channel = CondaChannel.from_url(x.channel).name

            if len(fv) >= 2:
                return f" - [white]{r.final_name:<30}[/white] [{color}]{version:<20}[/{color}] {fv[0]:<10} {fv[1]:<20} {channel}\n"
            else:
                return f" - [white]{r.final_name:<30}[/white] [{color}]{version:<20}[/{color}] {fv[0]:<20} {channel}\n"

        for r in self.requirements["build"]:
            s += spec_format(r)
        s += "Host:\n"
        for r in self.requirements["host"]:
            s += spec_format(r)
        s += "Run:\n"
        for r in self.requirements["run"]:
            s += spec_format(r)
        return s

    def propagate_run_exports(self, env, pkg_cache):
        # find all run exports
        collected_run_exports = []
        config_pins = self.conda_build_config.get("pin_run_as_build", {})
        for s in self.requirements[env]:
            if s.is_transitive_dependency:
                continue
            if s.name in self.sections["build"].get("ignore_run_exports", []):
                continue
            if hasattr(s, "final_version"):
                final_triple = (
                    f"{s.final_name}-{s.final_version[0]}-{s.final_version[1]}"
                )
            else:
                console.print(f"[red]{s} has no final version")
                continue

            if s.name.replace("-", "_") in config_pins:
                s.run_exports_info = {
                    "weak": [
                        f"{s.final_name} {apply_pin_expressions(s.final_version[0], **config_pins[s.name.replace('-', '_')])}"
                    ]
                }
                collected_run_exports.append(s.run_exports_info)
            else:
                path = Path(pkg_cache).joinpath(
                    final_triple, "info", "run_exports.json",
                )
                if path.exists():
                    with open(path) as fi:
                        run_exports_info = json.load(fi)
                        s.run_exports_info = run_exports_info
                        collected_run_exports.append(run_exports_info)
                else:
                    s.run_exports_info = None

        def append_or_replace(env, spec):
            spec = CondaBuildSpec(spec)
            name = spec.name
            spec.from_run_export = True
            for idx, r in enumerate(self.requirements[env]):
                if r.final_name == name:
                    self.requirements[env][idx] = spec
                    return
            self.requirements[env].append(spec)

        if env == "build":
            for rex in collected_run_exports:
                if not self.noarch:
                    for r in rex.get("strong", []):
                        append_or_replace("host", r)
                        append_or_replace("run", r)
                    for r in rex.get("weak", []):
                        append_or_replace("host", r)
                    for r in rex.get("strong_constrains", []):
                        append_or_replace("run_constrained", r)

        if env == "host":
            for rex in collected_run_exports:
                if not self.noarch:
                    for r in rex.get("strong", []):
                        append_or_replace("run", r)
                    for r in rex.get("weak", []):
                        append_or_replace("run", r)
                    for r in rex.get("strong_constrains", []):
                        append_or_replace("run_constrained", r)
                    for r in rex.get("weak_constrains", []):
                        append_or_replace("run_constrained", r)
                else:
                    for r in rex.get("noarch", []):
                        append_or_replace("run", r)

    def _solve_env(self, env, all_outputs):
        if self.requirements.get(env):
            console.print(f"Finalizing [yellow]{env}[/yellow] for {self.name}")
            specs = self.requirements[env]
            for s in specs:
                if s.is_pin:
                    s.eval_pin_subpackage(all_outputs)
                if env == "run" and s.is_pin_compatible:
                    s.eval_pin_compatible(
                        self.requirements["build"], self.requirements["host"]
                    )

            # save finalized requirements in data for usage in metadata
            self.data["requirements"][env] = [s.final for s in self.requirements[env]]

            spec_map = {s.final_name: s for s in specs}
            specs = [str(x) for x in specs]

            if env in ("host", "run") and not self.config.subdirs_same:
                subdir = self.config.host_subdir
            else:
                subdir = self.config.build_subdir

            solver, pkg_cache = get_solver(subdir)
            t = solver.solve(specs, [pkg_cache])

            _, install_pkgs, _ = t.to_conda()
            for _, _, p in install_pkgs:
                p = json.loads(p)
                if p["name"] in spec_map:
                    spec_map[p["name"]].final_version = (
                        p["version"],
                        p["build_string"],
                    )
                    spec_map[p["name"]].channel = p["channel"]
                else:
                    cbs = CondaBuildSpec(f"{p['name']}")
                    cbs.is_transitive_dependency = True
                    cbs.final_version = (p["version"], p["build_string"])
                    cbs.channel = p["channel"]
                    self.requirements[env].append(cbs)

            self.transactions[env] = {
                "transaction": t,
                "pkg_cache": pkg_cache,
            }

            downloaded = t.fetch_extract_packages(
                solver.repos + list(solver.local_repos.values()),
            )
            if not downloaded:
                raise RuntimeError("Did not succeed in downloading packages.")

            if env in ("build", "host"):
                self.propagate_run_exports(env, self.transactions[env]["pkg_cache"])

    def set_final_build_id(self, meta):
        self.final_build_id = meta.build_id()
        # we need to evaluate run_exports pin_subpackage here...
        if self.sections["build"].get("run_exports"):
            run_exports_list = self.sections["build"]["run_exports"]
            for x in run_exports_list:
                if self.name.endswith("-static") and self.name.startswith(x.name):
                    # remove self-run-exports for static packages
                    run_exports_list[:] = []
                else:
                    x.eval_pin_subpackage([self])

                x.eval_pin_subpackage([self])
            run_exports_list[:] = [x.final for x in run_exports_list]
            self.data["build"]["run_exports"] = run_exports_list

    def finalize_solve(self, all_outputs):

        self._solve_env("build", all_outputs)
        self._solve_env("host", all_outputs)
        self._solve_env("run", all_outputs)

        # TODO figure out if we can avoid this?!
        if self.config.variant.get("python") is None:
            for r in self.requirements["build"] + self.requirements["host"]:
                if r.name == "python":
                    self.config.variant["python"] = r.final_version[0]

        if self.config.variant.get("python") is None:
            self.config.variant["python"] = ".".join(
                [str(v) for v in sys.version_info[:3]]
            )

        self.variant = self.config.variant



# Copyright (C) 2021, QuantStack
# SPDX-License-Identifier: BSD-3-Clause

import os
import sys
import re

from conda.models.match_spec import MatchSpec
from conda.gateways.disk.create import mkdir_p

import conda_build.environ
from conda_build import api
from conda_build.config import Config, get_channel_urls
from conda_build.cli.main_build import parse_args
from conda_build.index import update_index

from boa.core.solver import MambaSolver
from boa.core.utils import normalize_subdir
from mamba.utils import init_api_context

only_dot_or_digit_re = re.compile(r"^[\d\.]+$")

solver_map = {}


def _get_solver(channel_urls, subdir, output_folder):
    """Gets a solver from cache or creates a new one if needed."""
    subdir = normalize_subdir(subdir)

    if subdir in solver_map:
        solver = solver_map[subdir]
        solver.replace_channels()
    else:
        solver = MambaSolver(channel_urls, subdir, output_folder)
        solver_map[subdir] = solver

    return solver


def mamba_get_install_actions(
    prefix,
    specs,
    env,
    retries=0,
    subdir=None,
    verbose=True,
    debug=False,
    locking=True,
    bldpkgs_dirs=None,
    timeout=900,
    disable_pip=False,
    max_env_retry=3,
    output_folder=None,
    channel_urls=None,
):
    solver = _get_solver(channel_urls, subdir, output_folder)

    _specs = [MatchSpec(s) for s in specs]
    for idx, s in enumerate(_specs):
        if s.version:
            vspec = str(s.version)
            if re.match(only_dot_or_digit_re, vspec) and vspec.count(".") <= 1:
                n = s.conda_build_form()
                sn = n.split()
                sn[1] = vspec + ".*"
                _specs[idx] = MatchSpec(" ".join(sn))

    _specs = [s.conda_build_form() for s in _specs]

    solution = solver.solve_for_action(_specs, prefix)
    return solution


conda_build.environ.get_install_actions = mamba_get_install_actions


def prepare(**kwargs):
    """
    Prepare and configure the stage for mambabuild to run.

    The given **kwargs are passed to conda-build's Config which
    is the value returned by this function.
    """
    config = Config(**kwargs)
    config.channel_urls = get_channel_urls(kwargs)

    init_api_context()

    config.output_folder = os.path.abspath(config.output_folder)
    if not os.path.exists(config.output_folder):
        mkdir_p(config.output_folder)

    print(f"Updating build index: {(config.output_folder)}\n")
    update_index(config.output_folder, verbose=config.debug, threads=1)

    return config


def call_conda_build(action, config, **kwargs):
    """
    After having set up the stage for boa's mambabuild to
    use the mamba solver, we delegate the work of building
    the conda package back to conda-build.

    Args:
        action: "build" or "test"
        config: conda-build's Config

    Kwargs:
        additional keyword arguments are passed to conda-build

    Return:
        The result of conda-build's build: the built packages
    """
    recipe = config.recipe[0]

    if action == "test":
        result = api.test(recipe, config=config, **kwargs)
    elif action == "build":
        result = api.build(
            recipe,
            post=config.post,
            build_only=config.build_only,
            notest=config.notest,
            config=config,
            variants=config.variants,
            **kwargs,
        )
    else:
        raise ValueError("action should be 'build' or 'test', got: %r" % action)

    return result


def main():
    _, args = parse_args(sys.argv[1:])

    config = prepare(**args.__dict__)

    action = "test" if args.test else "build"

    call_conda_build(action, config)



# Copyright (C) 2021, QuantStack
# SPDX-License-Identifier: BSD-3-Clause

from typing import Union, Dict, Iterable, Any, Optional

try:
    from conda_build.metadata import default_structs, ARCH_MAP

    FIELDS = None
except ImportError:
    from conda_build.metadata import FIELDS, ARCH_MAP

    default_structs = None

from conda_build.utils import ensure_list

import hashlib
import os
import time
import sys
import re
import json
import copy


def get_package_version_pin(specs, name):
    for s in specs:
        x = s.split(" ")
        if x[0] == name and len(x) > 1:
            return x[1]
    return None


def build_string_from_metadata(metadata):
    if metadata.meta.get("build", {}).get("string"):
        build_str = metadata.get_value("build/string")
    else:
        res = []
        build_or_host = "host" if metadata.is_cross else "build"
        build_pkg_names = [ms.name for ms in metadata.ms_depends(build_or_host)]
        build_deps = metadata.meta.get("requirements", {}).get(build_or_host, [])
        # TODO: this is the bit that puts in strings like py27np111 in the filename.  It would be
        #    nice to get rid of this, since the hash supercedes that functionally, but not clear
        #    whether anyone's tools depend on this file naming right now.
        for s, names, places in (
            ("np", "numpy", 2),
            ("py", "python", 2),
            ("pl", "perl", 2),
            ("lua", "lua", 2),
            ("r", ("r", "r-base"), 2),
            ("mro", "mro-base", 3),
            ("mro", "mro-base_impl", 3),
        ):

            for ms in metadata.ms_depends("run"):
                for name in ensure_list(names):
                    if ms.name == name and name in build_pkg_names:
                        # only append numpy when it is actually pinned
                        if name == "numpy" and not metadata.numpy_xx:
                            continue
                        if metadata.noarch == name or (
                            metadata.get_value("build/noarch_python")
                            and name == "python"
                        ):
                            res.append(s)
                        else:
                            pkg_names = list(ensure_list(names))
                            pkg_names.extend(
                                [
                                    _n.replace("-", "_")
                                    for _n in ensure_list(names)
                                    if "-" in _n
                                ]
                            )
                            for _n in pkg_names:
                                variant_version = get_package_version_pin(
                                    build_deps, _n
                                ) or metadata.config.variant.get(
                                    _n.replace("-", "_"), ""
                                )
                                if variant_version:
                                    break
                            entry = "".join([s] + variant_version.split(".")[:places])
                            if entry not in res:
                                res.append(entry)

        features = ensure_list(metadata.get_value("build/features", []))
        if res:
            res.append("_")
        if features:
            res.extend(("_".join(features), "_"))
        res.append(
            "{0}".format(metadata.build_number() if metadata.build_number() else 0)
        )
        build_str = "".join(res)
    return build_str


class MetaData:

    path: str
    uses_new_style_compiler_activation: bool = False
    uses_vcs_in_meta: bool = False
    uses_vcs_in_build: bool = False
    build_is_host: bool = False
    is_cross: bool = True
    final: bool = True
    activate_build_script: bool = True

    numpy_xx: bool = False
    noarch: Optional[str] = None
    noarch_python: bool = False

    # TODO? What are the implications of this?!
    is_output: bool = False
    pin_depends: bool = False

    def __init__(self, path, output):
        self.config = output.config
        self.meta = output.data
        self.output = output

        if os.path.isdir(path):
            self.path = os.path.abspath(path)
        else:
            self.path = os.path.dirname(os.path.abspath(path))

        self._meta_name = "recipe.yaml"
        self._meta_path = os.path.join(self.path, self._meta_name)

        self.noarch = self.output.sections["build"].get("noarch", None)

    def get_section(self, key: str) -> Union[Dict, Iterable]:
        return self.output.sections[key]

    def skip(self):
        return self.output.skip()

    def get_value(self, in_key: str, default: Any = None, autotype=True) -> Any:
        if in_key.count("/") == 2:
            section, num, key = in_key.split("/")
        else:
            section, key = in_key.split("/")
            num = 0

        # conda-build compat
        if default_structs:
            if default is None and in_key in default_structs:
                default = default_structs[in_key]()
        elif FIELDS:
            if autotype and default is None and FIELDS.get(section, {}).get(key):
                default = FIELDS[section][key]()

        section = self.output.sections.get(section, {})
        if isinstance(section, list):
            return section[int(num)].get(key, default)
        else:
            return section.get(key, default)

    @property
    def source_provided(self):
        return not bool(self.meta.get("source")) or (
            os.path.isdir(self.config.work_dir)
            and len(os.listdir(self.config.work_dir)) > 0
        )

    def ms_depends(self, typ="run"):
        names = ("python", "numpy", "perl", "lua")
        name_ver_list = [
            (name, self.config.variant[name])
            for name in names
            if self.config.variant.get(name)
        ]
        if self.config.variant.get("r_base"):
            # r is kept for legacy installations, r-base deprecates it.
            name_ver_list.extend(
                [
                    ("r", self.config.variant["r_base"]),
                    ("r-base", self.config.variant["r_base"]),
                ]
            )
        specs = []
        # for spec in ensure_list(self.get_value('requirements/' + typ, [])):
        for spec in self.get_dependencies(typ):
            if not spec:
                continue
            if spec.is_transitive_dependency and not spec.from_run_export:
                continue

            if spec.name == self.name():
                raise RuntimeError("%s cannot depend on itself" % self.name())

            for name, _ in name_ver_list:
                if spec.name == name:
                    if self.noarch:
                        continue

            for c in "=!@#$%^&*:;\"'\\|<>?/":
                if c in spec.name:
                    sys.exit(
                        "Error: bad character '%s' in package name "
                        "dependency '%s'" % (c, spec.name)
                    )

            parts = spec.splitted
            if len(parts) >= 2:
                if parts[1] in {">", ">=", "=", "==", "!=", "<", "<="}:
                    msg = (
                        "Error: bad character '%s' in package version "
                        "dependency '%s'" % (parts[1], spec.name)
                    )
                    if len(parts) >= 3:
                        msg += "\nPerhaps you meant '%s %s%s'" % (
                            spec.name,
                            parts[1],
                            parts[2],
                        )
                    sys.exit(msg)
            specs.append(spec)

        return specs

    def name(self, fail_ok=False):
        return self.output.name

    def version(self):
        return self.output.version

    def build_string(self):
        return self.output.build_string

    def build_number(self):
        return self.output.build_number

    def include_recipe(self):
        return self.get_value("build/include_recipe", True)

    def use_feature_map(self):
        return self.output.feature_map

    def build_features(self):
        m = self.use_feature_map()

        def truefalse(x):
            if x:
                return "1"
            else:
                return "0"

        return {"FEATURE_" + k.upper(): truefalse(v["activated"]) for k, v in m.items()}

    @property
    def meta_path(self):
        meta_path = self._meta_path or self.meta.get("extra", {}).get(
            "parent_recipe", {}
        ).get("path", "")
        if meta_path and os.path.basename(meta_path) != self._meta_name:
            meta_path = os.path.join(meta_path, self._meta_name)
        return meta_path

    def hash_dependencies(self):
        """With arbitrary pinning, we can't depend on the build string as done in
        build_string_from_metadata - there's just too much info.  Instead, we keep that as-is, to
        not be disruptive, but we add this extra hash, which is just a way of distinguishing files
        on disk.  The actual determination of dependencies is done in the repository metadata.

        This was revised in conda-build 3.1.0: hashing caused too many package
            rebuilds. We reduce the scope to include only the pins added by conda_build_config.yaml,
            and no longer hash files that contribute to the recipe.
        """
        hash_ = ""
        hashing_dependencies = self.get_hash_contents()
        if hashing_dependencies:
            hash_ = hashlib.sha1(
                json.dumps(hashing_dependencies, sort_keys=True).encode()
            )
            # save only the first HASH_LENGTH characters - should be more than
            #    enough, since these only need to be unique within one version
            # plus one is for the h - zero pad on the front, trim to match HASH_LENGTH
            hash_ = "h{0}".format(hash_.hexdigest())[: self.config.hash_length + 1]
        return hash_

    def build_id(self):
        manual_build_string = self.get_value("build/string")
        if manual_build_string:
            out = manual_build_string
        else:
            # default; build/string not set or uses PKG_HASH variable, so we should fill in the hash
            out = build_string_from_metadata(self)
            if self.config.filename_hashing and self.final:
                hash_ = self.hash_dependencies()
                if not re.findall("h[0-9a-f]{%s}" % self.config.hash_length, out):
                    ret = out.rsplit("_", 1)
                    try:
                        int(ret[0])
                        out = "_".join((hash_, str(ret[0]))) if hash_ else str(ret[0])
                    except ValueError:
                        out = ret[0] + hash_
                    if len(ret) > 1:
                        out = "_".join([out] + ret[1:])
                else:
                    out = re.sub("h[0-9a-f]{%s}" % self.config.hash_length, hash_, out)
        return out

    def dist(self):
        return "%s-%s-%s" % (self.name(), self.version(), self.build_id())

    def always_include_files(self):
        return []

    def binary_relocation(self):
        return True

    def get_dependencies(self, which):
        deps = self.output.requirements[which]
        # print(deps)
        # for feat, used in self.use_feature_map().items():
        #     if used:
        #         fdeps = used.get('requirements')
        #         if fdeps:
        #             fdeps = fdeps.get(which, [])
        #         deps.extend(fdeps)
        return deps

    def get_hash_contents(self):
        """
        # A hash will be added if all of these are true for any dependency:
        #
        # 1. package is an explicit dependency in build, host, or run deps
        # 2. package has a matching entry in conda_build_config.yaml which is a pin to a specific
        #    version, not a lower bound
        # 3. that package is not ignored by ignore_version
        #
        # The hash is computed based on the pinning value, NOT the build
        #    dependency build string. This means hashes won't change as often,
        #    but it also means that if run_exports is overly permissive,
        #    software may break more often.
        #
        # A hash will also ALWAYS be added when a compiler package is a build
        #    or host dependency. Reasoning for that is that the compiler
        #    package represents compiler flags and other things that can and do
        #    dramatically change compatibility. It is much more risky to drop
        #    this info (by dropping the hash) than it is for other software.

        # used variables - anything with a value in conda_build_config.yaml that applies to this
        #    recipe.  Includes compiler if compiler jinja2 function is used.
        """

        # trim_build_only_deps(self, dependencies)
        dependencies = (
            self.get_dependencies("build")
            + self.get_dependencies("host")
            # self.output.requirements["build"] + self.output.requirements["host"]
        )
        dependencies = {x.name for x in dependencies}
        # filter out ignored versions
        build_string_excludes = ["python", "r_base", "perl", "lua", "target_platform"]
        build_string_excludes.extend(
            ensure_list(self.config.variant.get("ignore_version", []))
        )

        # TODO
        # if 'numpy' in dependencies:
        #     pin_compatible, not_xx = self.uses_numpy_pin_compatible_without_xx
        #     # numpy_xx means it is accounted for in the build string, with npXYY
        #     # if not pin_compatible, then we don't care about the usage, and omit it from the hash.
        #     if self.numpy_xx or (pin_compatible and not not_xx):
        #         build_string_excludes.append('numpy')
        # always exclude older stuff that's always in the build string (py, np, pl, r, lua)
        if build_string_excludes:
            exclude_pattern = re.compile(
                "|".join("{}[\\s$]?.*".format(exc) for exc in build_string_excludes)
            )
            filtered_deps = []
            for req in dependencies:
                if exclude_pattern.match(req):
                    continue
                if req in self.config.variant:
                    if " " in self.config.variant[req]:
                        continue
                filtered_deps.append(req)

        take_keys = set(self.config.variant.keys())
        if "python" in take_keys and "python" not in dependencies:
            take_keys.remove("python")

        # retrieve values - this dictionary is what makes up the hash.
        return {key: self.config.variant[key] for key in take_keys}

    def info_index(self):
        arch = (
            "noarch" if self.config.target_subdir == "noarch" else self.config.host_arch
        )
        d = dict(
            name=self.name(),
            version=self.version(),
            build=self.build_id(),
            build_number=self.build_number() if self.build_number() else 0,
            platform=self.config.platform
            if (self.config.platform != "noarch" and arch != "noarch")
            else None,
            arch=ARCH_MAP.get(arch, arch),
            subdir=self.config.target_subdir,
            depends=sorted(
                " ".join(ms.final.split(" ")[:2]) for ms in self.ms_depends()
            ),
            timestamp=int(time.time() * 1000),
        )
        for key in ("license", "license_family"):
            value = self.get_value("about/" + key)
            if value:
                d[key] = value

        preferred_env = self.get_value("build/preferred_env")
        if preferred_env:
            d["preferred_env"] = preferred_env

        # conda 4.4+ optional dependencies
        constrains = ensure_list(self.get_value("requirements/run_constrained"))
        # filter None values
        constrains = [str(v) for v in constrains if v]
        if constrains:
            d["constrains"] = constrains

        if self.get_value("build/features"):
            d["features"] = " ".join(self.get_value("build/features"))
        if self.get_value("build/track_features"):
            d["track_features"] = " ".join(self.get_value("build/track_features"))
        if self.get_value("build/provides_features"):
            d["provides_features"] = self.get_value("build/provides_features")
        if self.get_value("build/requires_features"):
            d["requires_features"] = self.get_value("build/requires_features")
        if self.noarch:
            d["platform"] = d["arch"] = None
            d["subdir"] = "noarch"
            # These are new-style noarch settings.  the self.noarch setting can be True in 2 ways:
            #    if noarch: True or if noarch_python: True.  This is disambiguation.
            build_noarch = self.get_value("build/noarch")
            if build_noarch:
                d["noarch"] = build_noarch

        # TODO
        # if self.is_app():
        #     d.update(self.app_meta())
        return d

    def ignore_prefix_files(self):
        return self.get_value("build/ignore_prefix_files")

    def binary_has_prefix_files(self):
        return self.get_value("build/binary_has_prefix_files", [])

    def has_prefix_files(self):
        return self.get_value("build/has_prefix_files", [])

    def copy(self):
        # delete transactions as we can't copy them
        # TODO find a better way ...
        self.output.transactions = None
        new = copy.deepcopy(self)
        return new

    def get_test_deps(self, py_files, pl_files, lua_files, r_files):
        specs = ["%s %s %s" % (self.name(), self.version(), self.build_id())]

        # add packages listed in the run environment and test/requires
        specs.extend(ms.final for ms in self.ms_depends("run"))
        specs += self.get_value("test/requires", [])
        spec_names = set((s.name for s in self.ms_depends("run")))

        if py_files and "python" not in spec_names:
            # as the tests are run by python, ensure that python is installed.
            specs += ["python"]
        if pl_files and "perl" not in spec_names:
            # as the tests are run by perl, we need to specify it
            specs += ["perl"]
        if lua_files and "lua" not in spec_names:
            # not sure how this shakes out
            specs += ["lua"]
        if r_files and not any(s in ("r-base", "mro-base") for s in spec_names):
            # not sure how this shakes out
            specs += ["r-base"]

        # What is this?!
        # specs.extend(utils.ensure_list(self.config.extra_deps))
        return specs
